/****************************************************************
 * mempak.c
 * Last change 12-22-2013 by Osman Celimli
 ****************************************************************/
#define _CRTDBG_MAP_ALLOC
#include <stdlib.h>
#include <crtdbg.h>

#include <stdio.h>
#include <string.h>
#include <windows.h>

#include "useful.h"
#include "options.h"
/****************************************************************
 * MEMPAK Flash Structure, starting from the base of the flash: 
 *	$00-$05	"MEMPAK" In ASCII
 *	$06	Number of Files (Lo,Hi)
 *	$08+	Directory entries in the following format:
 *		$00-$01 ASCII File Type (Default is 'FX')
 *		$02-$04	Starting address of the File Data (Lo,Md,Hi)
 *		$05-$07	Size of the File in Bytes (Lo,Md,Hi)
 *	The remaining memory following the directory entries is
 *	used for file data.
 ****************************************************************/
#define HEADER_LENGTH		8
#define HEADER_MAGIC		"MEMPAK"
#define HEADER_OFF_MAGIC	0
#define HEADER_OFF_SIZE		6

#define	DIR_LENGTH			8
#define	DIR_OFF_TYPE		0
#define	DIR_OFF_START		2
#define	DIR_OFF_SIZE		5

#define	MAX_MEMPAK_SIZE		1024 * 1024 * 16
/****************************************************************
 * Allocation Settings
 ****************************************************************/
#define	INIT_PACK_SIZE		2048 * sizeof(char)
#define INIT_DIR_SIZE		256 * sizeof(char)
#define INIT_OFFS_SIZE		64 * sizeof(int)

#define INIT_SCRIPT_SIZE	256 * sizeof(char)
/****************************************************************
 * Options
 ****************************************************************/
typedef int (*optionSetup)(char *);
typedef struct optionPair
{
	char szName[3];
	optionSetup pSetup;
} optionPair;

optionPair aOptions[]={
	{"-f", f_openScript},
	{"-e", e_setType},
	{"-a", a_genAdjTable},
	{"-b", b_setBitTable},
	{"-m", m_setBitMask}
};
#define NUM_OPTIONS (sizeof(aOptions) / sizeof(optionPair))

typedef unsigned char (*byteOp)(unsigned char);
byteOp aByteOps[]={
	a_adjVoice,
	b_bitBobble,
	m_maskBits
};
#define NUM_BYTEOPS (sizeof(aByteOps) / sizeof(byteOp))
/****************************************************************
 * Important Globals
 ****************************************************************/
unsigned int uiNumFiles = 0;
unsigned int uiCurOff = 0;
unsigned int uiDirOff = 0;

unsigned char *pDir = NULL;
unsigned int uiDirLen,uiDirMemLen;
unsigned char *pFilePack = NULL;
unsigned int uiPackLen,uiPackMemLen;

unsigned int *pauiOffs = NULL;
unsigned int uiOffMemLen;
/****************************************************************
 * Text Tables
 ****************************************************************/
char ERR_Usage[]=
	"Usage: mempak outfile.bin [options] [source files]\n"
	"\n"
	"Supported Options:\n"
	"-f infile.x  : Pull more MEMPAK arguments from infile.x\n"
	"-e ex        : Set file type to 'ex' (default is 'FX').\n"
	"-a sweep.csv : Use amplitude sweep for 8-Bit DAC bias & step correction.\n"
	"-b bittable  : Rearrange the bits in each byte (ex. '32107654').\n"
	"-m bitmask   : Mask the bits of each byte (decimal, $hex, or %bin).\n"
	"\n"
	"Source data and options may be interleaved, for example :\n"
	"    mempak outfile.bin -b 01234567 source_a.bin -b 32107654 source_b.bin\n"
	"would yield a binary where the first file has the bits of each byte\n"
	"in reverse, but the second has the nibbles of each byte swapped.\n";

char ERR_NoMainMem[]= "ERROR- Out of memory\n";
char ERR_NoIncludes[]= "ERROR- No include files provided\n";
char ERR_EarlyEnd[]= "ERROR- Unexpected end of source data in : ";
char ERR_TooMuch[]= "ERROR- Binary exceeds the 16MB MEMPAK limit.\n";

char ERR_NoOpen[]= "ERROR- Can't open source file : ";
char ERR_NoDest[]= "ERROR- Can't write to destination file : ";

char ERR_BadOption[]= "ERROR- Invalid option : ";
char ERR_BadArgs[]= "ERROR- Missing or invalid argument for option : ";
char ERR_BadScript[]= "ERROR- Malformed packing script\n";

char REM_Done[]= "MEMPAK Binary saved to target : ";
/****************************************************************
 * int procPak(char *pszScript)
 * Process the supplied command script or subscript, containing
 * only options and included files separated by spaces ' '.
 ****************************************************************/
int procPak(char *pszScript)
{
	char *pszCurTok,*pszSubTok;
	unsigned int uiScriptLen,uiTokenStart,uiTokenEnd;
	unsigned int uiX,uiY;

	FILE *pCurFile;
	unsigned char ucByte;
	unsigned int uiCurFile,uiCurSize;
	int iFail;

	uiScriptLen = strlen(pszScript);
	uiTokenStart = 0;
	uiTokenEnd = 0;

	/**
	 * Scan the script until our current token end marker has
	 * bumped up against the null terminator.
	 */
	while(uiTokenEnd < uiScriptLen)
	{
		/**
		 * Find the end of the current token...
		 */
		uiX = uiTokenStart;
		while((pszScript[uiX] != ' ') && (pszScript[uiX] != '\0')) uiX++;
		/**
		 * ...and make sure it has a length over zero, otherwise we
		 * may be parsing a malformed script.
		 */
		if((uiTokenEnd = uiX) == uiTokenStart)
		{
			fprintf(stderr, "%s", ERR_BadScript);
			return -1;
		}
		pszScript[uiTokenEnd] = '\0';

		pszCurTok = pszScript + uiTokenStart;
		pszSubTok = (uiTokenEnd < uiScriptLen) ? (pszScript + uiTokenEnd + 1) : NULL;

		/**
		 * Check to see if the current token is an option, pulling
		 * in its parameters if needed.
		 */
		if(pszCurTok[0] == '-')
		{
			for(uiX = 0; uiX < NUM_OPTIONS; uiX++)
			{
				if(strcmp(aOptions[uiX].szName, pszCurTok) == 0) break;
			}
			if(uiX < NUM_OPTIONS)
			{
				if(pszSubTok == NULL)
				{
					fprintf(stderr, "%s%s\n", ERR_BadArgs, pszCurTok);
					return -1;
				}

				uiTokenStart = uiTokenEnd + 1;
				uiY = uiTokenStart;
				while((pszScript[uiY] != ' ') && (pszScript[uiY] != '\0')) uiY++;

				uiTokenEnd = uiY;
				pszScript[uiTokenEnd] = '\0';

				if((*(aOptions[uiX].pSetup))(pszSubTok)) return -1;
			}
			else
			{
				fprintf(stderr, "%s%s\n", ERR_BadOption, pszCurTok);
				return -1;
			}
			uiTokenStart = uiTokenEnd + 1;
			continue;
		}

		/**
		 * Okay, it's got to be a file. Let's check its length and
		 * then see if we need to adjust any of our buffers before
		 * stuffing it into our binary.
		 */
		if(NULL == (pCurFile = fopen(pszCurTok, "rb")))
		{
			fprintf(stderr,"%s%s\n", ERR_NoOpen, pszCurTok);
			return -1;
		}
		fseek(pCurFile, 0, SEEK_END);
		uiCurSize = ftell(pCurFile);
		rewind(pCurFile);

		uiCurFile = uiNumFiles;
		uiDirLen += DIR_LENGTH;
		uiPackLen += uiCurSize;
		uiOffMemLen += sizeof(unsigned int);
		uiNumFiles++;

		iFail = tealloc(&pDir,
			&uiDirMemLen, uiDirLen);
		iFail |= tealloc(&pFilePack,
			&uiPackMemLen, uiPackLen);
		iFail |= tealloc(&pauiOffs,
			&uiOffMemLen, (uiNumFiles * sizeof(unsigned int)));
		if(iFail)
		{
			fprintf(stderr, "%s", ERR_NoMainMem);
			fclose(pCurFile);
			return -1;
		}

		/**
		 * Read in the file, stamp its type and size in the directory,
		 * then mark its offset for when we do the final build later on.
		 */
		if(uiCurSize != fread(pFilePack + uiCurOff, sizeof(unsigned char), uiCurSize, pCurFile))
		{
			fprintf(stderr, "%s%s\n", ERR_EarlyEnd, pszCurTok);
			fclose(pCurFile);
			return -1;
		}
		fclose(pCurFile);

		pDir[uiDirOff + DIR_OFF_TYPE + 0] = gszFileType[0];
		pDir[uiDirOff + DIR_OFF_TYPE + 1] = gszFileType[1];

		pDir[uiDirOff + DIR_OFF_SIZE + 0] = uiCurSize & 0xFF;
		pDir[uiDirOff + DIR_OFF_SIZE + 1] = (uiCurSize >> 8) & 0xFF;
		pDir[uiDirOff + DIR_OFF_SIZE + 2] = (uiCurSize >> 16) & 0xFF;

		pauiOffs[uiCurFile] = uiCurOff;

		/**
		 * Don't forget to perform our adjustments on the individual bytes
		 * of the file before moving on to the next token.
		 * Adjustments must be performed in the order they're placed in
		 * the byte operations table, passing the result of the last
		 * operation to the next one.
		 */
		for(uiX = uiCurOff; uiX < (uiCurOff + uiCurSize); uiX++)
		{
			ucByte = pFilePack[uiX];
			for(uiY = 0; uiY < NUM_BYTEOPS; uiY++)
			{
				ucByte = (*(aByteOps[uiY]))(ucByte);
			}
			pFilePack[uiX] = ucByte;
		}
		uiDirOff += DIR_LENGTH;
		uiCurOff += uiCurSize;
		pCurFile = NULL;

		/**
		 * Wait a minute... have we exceeded the capacity of
		 * the MEMPAK format?
		 */
		if((HEADER_LENGTH + uiDirLen + uiPackLen) > MAX_MEMPAK_SIZE)
		{
			fprintf(stderr, "%s", ERR_TooMuch);
			return -1;
		}
		uiTokenStart = uiTokenEnd + 1;
	}
	return 0;
}

/****************************************************************
 * int main(int argc, char **argv)
 * mempak...
 ****************************************************************/
int main(int argc, char **argv)
{
	/* Vars */
	FILE *pDest = NULL;
	char *pszScript = NULL;
	unsigned int uiScriptLen,uiScriptMemLen;

	unsigned char *pFinalPack = NULL;
	unsigned int uiFinalSize;

	unsigned int uiX,uiY,uiZ;

	/* Look for leaks if we're debuggin' */
#ifdef _DEBUG
	int tMemDbg;
	_CrtSetReportMode( _CRT_ERROR, _CRTDBG_MODE_DEBUG );
	tMemDbg = _CrtSetDbgFlag( _CRTDBG_REPORT_FLAG );
	tMemDbg |= _CRTDBG_CHECK_ALWAYS_DF;
	tMemDbg |= _CRTDBG_CHECK_CRT_DF;
	tMemDbg |= _CRTDBG_DELAY_FREE_MEM_DF;
	tMemDbg |= _CRTDBG_LEAK_CHECK_DF;
	_CrtSetDbgFlag( tMemDbg );
#endif

	/**
	 * Show usage for no arguments, otherwise try to open the first
	 * argmument for writing as our destination file.
	 */
	if(argc < 2)
	{
		fprintf(stderr, "%s", ERR_Usage);
		return 0;
	}
	if(NULL == (pDest = fopen(argv[1],"wb")))
	{
		fprintf(stderr, "%s%s\n", ERR_NoDest, argv[1]);
		return 0;
	}

	/**
	 * All arguments after the first are considered part of the base
	 * include script, which may be supplemented by other scripts defined
	 * within it or themselves (recursive inclusion). However, since
	 * the script format is just the individual files and options
	 * separated by spaces (' '), we need to do some conversion first.
	 */
	pszScript = malloc(INIT_SCRIPT_SIZE);
	uiScriptMemLen = INIT_SCRIPT_SIZE;
	uiScriptLen = 0;

	if(pszScript == NULL)
	{
		fprintf(stderr, "%s", ERR_NoMainMem);
		goto cleanup;
	}
	pszScript[0] = '\0';

	uiZ = argc;
	for(uiX = 2; uiX < uiZ; uiX++)
	{
		if(uiScriptLen > 0) pszScript[uiScriptLen - 1] = ' ';
		uiY = strlen(argv[uiX]) + 1;

		if(tealloc(&pszScript, &uiScriptMemLen, (uiScriptLen + uiY)))
		{
			fprintf(stderr, "%s", ERR_NoMainMem);
			goto cleanup;
		}

		strcpy(pszScript + uiScriptLen, argv[uiX]);
		uiScriptLen += uiY;
	}

	/**
	 * Prep the directory, offset, and file data buffers...
	 */
	pDir = malloc(INIT_DIR_SIZE);
	uiDirMemLen = INIT_DIR_SIZE;
	uiDirLen = 0;

	pFilePack = malloc(INIT_PACK_SIZE);
	uiPackMemLen = INIT_PACK_SIZE;
	uiPackLen = 0;

	pauiOffs = malloc(INIT_OFFS_SIZE);
	uiOffMemLen = INIT_OFFS_SIZE;

	if((pDir == NULL) || (pauiOffs == NULL) || (pFilePack == NULL))
	{
		fprintf(stderr, "%s", ERR_NoMainMem);
		goto cleanup;
	}
	
	/**
	 * ...then feed our script into the pack processor. But
	 * don't forget to check if we actually got any include
	 * files before building final MEMPAK binary.
	 */
	if(procPak(pszScript)) goto cleanup;
	if(uiNumFiles == 0)
	{
		fprintf(stderr, "%s", ERR_NoIncludes);
		goto cleanup;
	}

	/**
	 * We can now build the final MEMPAK binary.
	 * Populate the offset fields in the directory structure, then
	 * combine the directory and file regions with a header.
	 */
	for(uiX = 0; uiX < uiNumFiles; uiX++)
	{
		uiY = HEADER_LENGTH + uiDirLen + pauiOffs[uiX];

		pDir[(DIR_LENGTH * uiX) + DIR_OFF_START + 0] = uiY & 0xFF;
		pDir[(DIR_LENGTH * uiX) + DIR_OFF_START + 1] = (uiY >> 8) & 0xFF;
		pDir[(DIR_LENGTH * uiX) + DIR_OFF_START + 2] = (uiY >> 16) & 0xFF;
	}

	uiFinalSize = HEADER_LENGTH + uiDirLen + uiPackLen;
	pFinalPack = malloc(uiFinalSize);
	if(pFinalPack == NULL)
	{
		fprintf(stderr, "%s", ERR_NoMainMem);
		goto cleanup;
	}

	strcpy(pFinalPack, HEADER_MAGIC);
	pFinalPack[HEADER_OFF_SIZE + 0] = uiNumFiles & 0xFF;
	pFinalPack[HEADER_OFF_SIZE + 1] = (uiNumFiles >> 8) & 0xFF;

	memcpy(pFinalPack + HEADER_LENGTH, pDir, uiDirLen);
	memcpy(pFinalPack + HEADER_LENGTH + uiDirLen, pFilePack, uiPackLen);

	if(uiFinalSize != fwrite(pFinalPack, sizeof(unsigned char), uiFinalSize, pDest))
	{
		fprintf(stderr, "%s%s\n", ERR_NoDest, argv[1]);
	}
	else
	{
		fprintf(stdout, "%s%s\n", REM_Done, argv[1]);
	}

cleanup:
	f_closeTree();
	if(pDest != NULL) fclose(pDest);
	if(pszScript != NULL) free(pszScript);

	if(pauiOffs != NULL) free(pauiOffs);
	if(pDir != NULL) free(pDir);
	if(pFilePack != NULL) free(pFilePack);
	if(pFinalPack != NULL) free(pFinalPack);
	return 0;
}
